from enfobench.dataset import Dataset
from enfobench.evaluation import AuthorInfo, ForecasterType, Model, ModelInfo

__all__ = [
    "Model",
    "ModelInfo",
    "AuthorInfo",
    "ForecasterType",
    "Dataset",
]



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Protocol

import pandas as pd


class ForecasterType(str, Enum):
    point = "point"
    quantile = "quantile"
    density = "density"
    ensemble = "ensemble"


@dataclass
class AuthorInfo:
    """Author information.

    Attributes:
        name: Name of the author.
        email: Email of the author.
    """

    name: str
    email: str | None = None


@dataclass
class ModelInfo:
    """Model information.

    Attributes:
        name: Name of the model.
        authors: List of authors.
        type: Type of the model.
        params: Parameters of the model.
    """

    name: str
    authors: list[AuthorInfo]
    type: ForecasterType  # noqa: A003
    params: dict[str, Any] = field(default_factory=dict)


class Model(Protocol):
    def info(self) -> ModelInfo:
        ...

    def forecast(
        self,
        horizon: int,
        history: pd.DataFrame,
        past_covariates: pd.DataFrame | None = None,
        future_covariates: pd.DataFrame | None = None,
        level: list[int] | None = None,
        **kwargs,
    ) -> pd.DataFrame:
        ...



import numpy as np
import pandas as pd
import pytest

from enfobench import AuthorInfo, ForecasterType, ModelInfo
from enfobench.evaluation.utils import create_forecast_index


class TestModel:
    def __init__(self, param1: int):
        self.param1 = param1

    def info(self) -> ModelInfo:
        return ModelInfo(
            name="TestModel",
            authors=[
                AuthorInfo("Author 1", "author-1@institution.org"),
            ],
            type=ForecasterType.point,
            params={
                "param1": 1,
            },
        )

    def forecast(
        self,
        horizon: int,
        history,
        past_covariates=None,  # noqa: ARG002
        future_covariates=None,  # noqa: ARG002
        level=None,  # noqa: ARG002
        **kwargs,  # noqa: ARG002
    ):
        index = create_forecast_index(history, horizon)
        y_hat = np.full(horizon, fill_value=history["y"].mean()) + self.param1
        return pd.DataFrame(
            index=index,
            data={
                "yhat": y_hat,
            },
        )


@pytest.fixture(scope="function")
def model():
    return TestModel(1)


@pytest.fixture(scope="session")
def target() -> pd.DataFrame:
    index = pd.date_range("2020-01-01", "2020-02-01", freq="30T")
    y = pd.DataFrame(
        index=index,
        data={
            "y": np.random.random(len(index)),
        },
    )
    return y


@pytest.fixture(scope="session")
def covariates() -> pd.DataFrame:
    index = pd.date_range("2020-01-01", "2020-02-01", freq="H")
    df = pd.DataFrame(
        index=index,
        data={
            "cov_1": np.random.random(len(index)),
            "cov_2": np.random.random(len(index)),
        },
    )
    return df


@pytest.fixture(scope="session")
def external_forecasts() -> pd.DataFrame:
    cutoff_dates = pd.date_range("2020-01-01", "2020-02-01", freq="D")
    forecasts = []

    for cutoff_date in cutoff_dates:
        index = pd.date_range(cutoff_date + pd.Timedelta(hours=1), cutoff_date + pd.Timedelta(days=7), freq="H")
        forecast = pd.DataFrame(
            data={
                "timestamp": index,
                "forecast_1": np.random.random(len(index)),
                "forecast_2": np.random.random(len(index)),
            },
        )
        forecast["cutoff_date"] = cutoff_date
        forecasts.append(forecast)

    return pd.concat(forecasts, ignore_index=True)


@pytest.fixture(scope="session")
def clean_forecasts() -> pd.DataFrame:
    cutoff_dates = pd.date_range("2020-01-01", "2021-01-01", freq="D")

    forecasts = []
    for cutoff_date in cutoff_dates:
        index = pd.date_range(cutoff_date + pd.Timedelta(hours=1), cutoff_date + pd.Timedelta(hours=25), freq="H")
        forecast = pd.DataFrame(
            data={
                "timestamp": index,
                "yhat": np.random.random(len(index)),
                "y": np.random.random(len(index)),
            },
        )
        forecast["cutoff_date"] = cutoff_date
        forecasts.append(forecast)

    forecast_df = pd.concat(forecasts, ignore_index=True)
    assert not forecast_df.isna().any(axis=None)
    return forecast_df


@pytest.fixture(scope="session")
def forecasts_with_missing_values() -> pd.DataFrame:
    cutoff_dates = pd.date_range("2020-01-01", "2021-01-01", freq="D")

    forecasts = []
    for cutoff_date in cutoff_dates:
        index = pd.date_range(cutoff_date + pd.Timedelta(hours=1), cutoff_date + pd.Timedelta(hours=25), freq="H")
        forecast = pd.DataFrame(
            data={
                "timestamp": index,
                "yhat": np.random.random(len(index)),
                "y": np.random.random(len(index)),
            },
        )
        forecast["cutoff_date"] = cutoff_date
        forecasts.append(forecast)

    forecast_df = pd.concat(forecasts, ignore_index=True)
    forecast_df.loc[forecast_df["y"] <= 0.3, "y"] = np.nan
    assert forecast_df.isna().any(axis=None)
    return forecast_df



from datetime import timedelta

import pandas as pd


def steps_in_horizon(horizon: pd.Timedelta, freq: str) -> int:
    """Return the number of steps in a given horizon.

    Args:
        horizon: The horizon to be split into steps.
        freq: The frequency of the horizon.

    Returns:
        The number of steps in the horizon.
    """
    freq = "1" + freq if not freq[0].isdigit() else freq
    periods = horizon / pd.Timedelta(freq)
    if not periods.is_integer():
        msg = f"Horizon {horizon} is not a multiple of the frequency {freq}"
        raise ValueError(msg)
    return int(periods)


def periods_in_duration(target: pd.DatetimeIndex, duration: timedelta | pd.Timedelta | str) -> int:
    if isinstance(duration, timedelta):
        duration = pd.Timedelta(duration)
    elif isinstance(duration, str):
        duration = pd.Timedelta(duration)

    first_delta = target[1] - target[0]
    last_delta = target[-1] - target[-2]
    assert first_delta == last_delta, "Season length is not constant"

    periods = duration / first_delta
    assert periods.is_integer(), "Season length is not a multiple of the frequency"
    return int(periods)


def create_forecast_index(history: pd.DataFrame, horizon: int) -> pd.DatetimeIndex:
    """Create time index for a forecast horizon.

    Args:
        history: The history of the time series.
        horizon: The forecast horizon.

    Returns:
        The time index for the forecast horizon.
    """
    last_date = history.index[-1]
    inferred_freq = history.index.inferred_freq
    freq = "1" + inferred_freq if not inferred_freq[0].isdigit() else inferred_freq
    return pd.date_range(
        start=last_date + pd.Timedelta(freq),
        periods=horizon,
        freq=freq,
    )


def generate_cutoff_dates(
    start_date: pd.Timestamp,
    end_date: pd.Timestamp,
    horizon: pd.Timedelta,
    step: pd.Timedelta,
) -> list[pd.Timestamp]:
    """Generate cutoff dates for cross-validation between two dates.

    The cutoff dates are separated by a fixed step size and the last cutoff date is a horizon away from the end date.

    Args:
        start_date: Start date of the time series.
        end_date: End date of the time series.
        horizon: Forecast horizon.
        step: Step size between cutoff dates.

    Examples
    --------
    >>> generate_cutoff_dates(
    ...     start_date=pd.Timestamp("2020-01-01"),
    ...     end_date=pd.Timestamp("2020-01-05"),
    ...     horizon=pd.Timedelta("2 days"),
    ...     step=pd.Timedelta("1 day"),
    ... )
    [
        Timestamp('2020-01-01 00:00:00'),
        Timestamp('2020-01-02 00:00:00'),
        Timestamp('2020-01-03 00:00:00'),
    ]
    """
    if horizon <= pd.Timedelta(0):
        msg = f"Horizon must be positive, got {horizon}."
        raise ValueError(msg)

    if step <= pd.Timedelta(0):
        msg = f"Step must be positive, got {step}."
        raise ValueError(msg)

    if horizon > end_date - start_date:
        msg = f"Horizon is longer than the evaluation period: {horizon} > {end_date - start_date}."
        raise ValueError(msg)

    if end_date <= start_date:
        msg = f"End date must be after the starting date, got {end_date} <= {start_date}."
        raise ValueError(msg)

    cutoff_dates = []

    cutoff = start_date
    while cutoff <= end_date - horizon:
        cutoff_dates.append(cutoff)
        cutoff += step

    if not cutoff_dates:
        msg = f"No cutoff dates between {start_date} and {end_date} with horizon {horizon} and step {step}."
        raise ValueError(msg)
    return cutoff_dates



class Dataset:
    def __init_subclass__(cls, **kwargs):
        """This throws a deprecation warning on subclassing."""
        msg = f"""
        {cls.__name__} is deprecated.
        Import it as 'from enfobench.dataset import Dataset' instead.
        """
        raise DeprecationWarning(msg)

    def __init__(self, *args, **kwargs):  # noqa
        """This throws a deprecation warning on initialization."""
        msg = f"""
        {self.__class__.__name__} is deprecated.
        Import it as 'from enfobench.dataset import Dataset' instead.
        """
        raise DeprecationWarning(msg)



from datetime import timedelta

import pandas as pd
import pytest

import enfobench.dataset.utils
from enfobench.evaluation import utils


@pytest.mark.parametrize(
    "horizon, freq, expected",
    [
        ("1 day", "15T", 96),
        ("1 day", "1H", 24),
        ("7 days", "1H", 7 * 24),
        ("1D", "1D", 1),
        ("1H", "1H", 1),
    ],
)
def test_steps_in_horizon(horizon, freq, expected):
    assert utils.steps_in_horizon(pd.Timedelta(horizon), freq) == expected


def test_steps_in_horizon_raises_with_non_multiple_horizon():
    with pytest.raises(ValueError):
        utils.steps_in_horizon(pd.Timedelta("36 minutes"), "15T")


def test_create_forecast_index(target):
    history = target
    horizon = 96
    last_date = history.index[-1]

    index = utils.create_forecast_index(history, horizon)

    assert isinstance(index, pd.DatetimeIndex)
    assert index.freq == target.index.freq
    assert len(index) == horizon
    assert all(idx > last_date for idx in index)


def test_create_perfect_forecasts_from_covariates(covariates):
    forecasts = enfobench.dataset.utils.create_perfect_forecasts_from_covariates(
        covariates,
        horizon=pd.Timedelta("7 days"),
        step=pd.Timedelta("1D"),
    )

    assert isinstance(forecasts, pd.DataFrame)
    assert "timestamp" in forecasts.columns
    assert "cutoff_date" in forecasts.columns
    assert all(col in forecasts.columns for col in covariates.columns)


@pytest.mark.parametrize(
    "freq, duration, expected",
    [
        ("15T", "15T", 1),
        ("15T", "1H", 4),
        ("15T", "1D", 96),
        ("15T", "2D", 2 * 96),
        ("15T", "1W", 7 * 96),
        ("15T", pd.Timedelta("15T"), 1),
        ("15T", pd.Timedelta("1H"), 4),
        ("15T", pd.Timedelta("1D"), 96),
        ("15T", pd.Timedelta("2D"), 2 * 96),
        ("15T", pd.Timedelta("1W"), 7 * 96),
        ("15T", timedelta(minutes=15), 1),
        ("15T", timedelta(hours=1), 4),
        ("15T", timedelta(days=1), 96),
        ("15T", timedelta(days=2), 2 * 96),
        ("15T", timedelta(weeks=1), 7 * 96),
        ("30T", "30T", 1),
        ("30T", "1H", 2),
        ("30T", "1D", 48),
        ("30T", "2D", 2 * 48),
        ("30T", "1W", 7 * 48),
        ("30T", pd.Timedelta("30T"), 1),
        ("30T", pd.Timedelta("1H"), 2),
        ("30T", pd.Timedelta("1D"), 48),
        ("30T", pd.Timedelta("2D"), 2 * 48),
        ("30T", pd.Timedelta("1W"), 7 * 48),
        ("30T", timedelta(minutes=30), 1),
        ("30T", timedelta(hours=1), 2),
        ("30T", timedelta(days=1), 48),
        ("30T", timedelta(days=2), 2 * 48),
        ("30T", timedelta(weeks=1), 7 * 48),
        ("1H", "1H", 1),
        ("1H", "1D", 24),
        ("1H", "2D", 2 * 24),
        ("1H", "1W", 7 * 24),
        ("1H", pd.Timedelta("1H"), 1),
        ("1H", pd.Timedelta("1D"), 24),
        ("1H", pd.Timedelta("2D"), 2 * 24),
        ("1H", pd.Timedelta("1W"), 7 * 24),
        ("1H", timedelta(hours=1), 1),
        ("1H", timedelta(days=1), 24),
        ("1H", timedelta(days=2), 2 * 24),
        ("1H", timedelta(weeks=1), 7 * 24),
    ],
)
def test_periods_in_duration(freq, duration, expected):
    target = pd.date_range(start="2020-01-01", end="2020-02-01", freq=freq)

    assert utils.periods_in_duration(target, duration) == expected



import pandas as pd
import pytest
from starlette.testclient import TestClient

from enfobench.dataset import Dataset
from enfobench.evaluation import (
    ForecastClient,
    cross_validate,
)
from enfobench.evaluation.evaluate import _compute_metric, _compute_metrics, evaluate_metrics
from enfobench.evaluation.metrics import mean_absolute_error, mean_squared_error
from enfobench.evaluation.server import server_factory
from enfobench.evaluation.utils import generate_cutoff_dates


@pytest.mark.parametrize(
    "horizon,step",
    [
        ("1 day", "1 day"),
        ("1 day", "1 hour"),
        ("1 day", "15 minutes"),
        ("7 days", "1 day"),
        ("7 days", "1 hour"),
        ("7 days", "15 minutes"),
    ],
)
def test_generate_cutoff_dates(horizon, step):
    start_date = pd.Timestamp("2020-01-01")
    end_date = pd.Timestamp("2020-01-10")
    horizon = pd.Timedelta(horizon)
    step = pd.Timedelta(step)

    cutoff_dates = generate_cutoff_dates(start_date, end_date, horizon, step)

    assert all(cutoff_date >= start_date for cutoff_date in cutoff_dates)
    assert all(cutoff_date + horizon <= end_date for cutoff_date in cutoff_dates)
    assert all(
        cutoff - previous_cutoff == step
        for cutoff, previous_cutoff in zip(cutoff_dates[1:], cutoff_dates[:-1], strict=True)
    )


def test_generate_cutoff_dates_raises_with_empty_time_series():
    with pytest.raises(ValueError):
        generate_cutoff_dates(
            start_date=pd.Timestamp("2020-01-01"),
            end_date=pd.Timestamp("2020-01-01"),
            horizon=pd.Timedelta("1 day"),
            step=pd.Timedelta("1 day"),
        )


def test_generate_cutoff_dates_raises_with_longer_horizon():
    with pytest.raises(ValueError):
        generate_cutoff_dates(
            start_date=pd.Timestamp("2020-01-01"),
            end_date=pd.Timestamp("2020-01-01"),
            horizon=pd.Timedelta("7 days"),
            step=pd.Timedelta("1 day"),
        )


def test_generate_cutoff_dates_raises_with_negative_horizon():
    with pytest.raises(ValueError):
        generate_cutoff_dates(
            start_date=pd.Timestamp("2020-01-01"),
            end_date=pd.Timestamp("2020-01-05"),
            horizon=pd.Timedelta("-1 day"),
            step=pd.Timedelta("1 day"),
        )


def test_generate_cutoff_dates_raises_with_negative_step():
    with pytest.raises(ValueError):
        generate_cutoff_dates(
            start_date=pd.Timestamp("2020-01-01"),
            end_date=pd.Timestamp("2020-01-05"),
            horizon=pd.Timedelta("1 day"),
            step=pd.Timedelta("-1 day"),
        )


def test_cross_validate_univariate_locally(model, target):
    forecasts = cross_validate(
        model=model,
        dataset=Dataset(target),
        start_date=pd.Timestamp("2020-01-10"),
        end_date=pd.Timestamp("2020-01-21"),
        horizon=pd.Timedelta("7 days"),
        step=pd.Timedelta("1 day"),
    )

    assert isinstance(forecasts, pd.DataFrame)
    assert "timestamp" in forecasts.columns
    assert "yhat" in forecasts.columns
    assert "cutoff_date" in forecasts.columns


def test_cross_validate_univariate_via_server(model, target):
    app = server_factory(model=model)
    test_client = TestClient(app)
    model = ForecastClient(client=test_client)

    forecasts = cross_validate(
        model=model,
        dataset=Dataset(target),
        start_date=pd.Timestamp("2020-01-10"),
        end_date=pd.Timestamp("2020-01-21"),
        horizon=pd.Timedelta("7 days"),
        step=pd.Timedelta("1 day"),
    )

    assert isinstance(forecasts, pd.DataFrame)
    assert "timestamp" in forecasts.columns
    assert "yhat" in forecasts.columns
    assert "cutoff_date" in forecasts.columns


def test_cross_validate_multivariate_locally(model, target, covariates, external_forecasts):
    forecasts = cross_validate(
        model=model,
        dataset=Dataset(target, covariates, external_forecasts),
        start_date=pd.Timestamp("2020-01-10"),
        end_date=pd.Timestamp("2020-01-21"),
        horizon=pd.Timedelta("7 days"),
        step=pd.Timedelta("1 day"),
    )

    assert isinstance(forecasts, pd.DataFrame)
    assert "timestamp" in forecasts.columns
    assert "yhat" in forecasts.columns
    assert "cutoff_date" in forecasts.columns


def test_cross_validate_multivariate_via_server(model, target, covariates, external_forecasts):
    app = server_factory(model=model)
    test_client = TestClient(app)
    model = ForecastClient(client=test_client)

    forecasts = cross_validate(
        model=model,
        dataset=Dataset(target, covariates, external_forecasts),
        start_date=pd.Timestamp("2020-01-10"),
        end_date=pd.Timestamp("2020-01-21"),
        horizon=pd.Timedelta("7 days"),
        step=pd.Timedelta("1 day"),
    )

    assert isinstance(forecasts, pd.DataFrame)
    assert "timestamp" in forecasts.columns
    assert "yhat" in forecasts.columns
    assert "cutoff_date" in forecasts.columns


def test_compute_metric(clean_forecasts):
    metric_value = _compute_metric(clean_forecasts, mean_absolute_error)

    assert 0 < metric_value < 1


def test_compute_metrics(clean_forecasts):
    metric_values = _compute_metrics(clean_forecasts, {"MAE": mean_absolute_error, "MSE": mean_squared_error})

    assert isinstance(metric_values, dict)
    assert 0 < metric_values["MAE"] < 1
    assert 0 < metric_values["MSE"] < 1


def test_compute_metric_on_forecast_with_missing_values_raises_error(forecasts_with_missing_values):
    with pytest.raises(ValueError):
        _compute_metric(forecasts_with_missing_values, mean_absolute_error)


def test_compute_metrics_on_forecast_with_missing_values_raises_error(forecasts_with_missing_values):
    with pytest.raises(ValueError):
        _compute_metrics(forecasts_with_missing_values, {"MAE": mean_absolute_error, "MSE": mean_squared_error})


def test_evaluate_metrics_on_clean_forecasts(clean_forecasts):
    metrics = evaluate_metrics(clean_forecasts, {"MAE": mean_absolute_error, "MSE": mean_squared_error})

    assert isinstance(metrics, pd.DataFrame)
    assert "MAE" in metrics.columns
    assert "MSE" in metrics.columns
    assert "weight" in metrics.columns
    assert len(metrics) == 1
    assert 0 < metrics["MAE"].iloc[0] < 1
    assert 0 < metrics["MSE"].iloc[0] < 1
    assert metrics["weight"].iloc[0] == 1


def test_evaluate_metrics_on_forecasts_with_missing_values(forecasts_with_missing_values):
    metrics = evaluate_metrics(forecasts_with_missing_values, {"MAE": mean_absolute_error, "MSE": mean_squared_error})

    assert isinstance(metrics, pd.DataFrame)
    assert "MAE" in metrics.columns
    assert "MSE" in metrics.columns
    assert "weight" in metrics.columns
    assert len(metrics) == 1
    assert 0 < metrics["MAE"].iloc[0] < 1
    assert 0 < metrics["MSE"].iloc[0] < 1
    assert pytest.approx(metrics["weight"].iloc[0], 0.1) == 1 - 0.3


def test_evaluate_metrics_on_clean_forecasts_grouped_by(clean_forecasts):
    metrics = evaluate_metrics(
        clean_forecasts,
        {"MAE": mean_absolute_error, "MSE": mean_squared_error},
        groupby="cutoff_date",
    )

    grouped_values = clean_forecasts["cutoff_date"].unique()

    assert isinstance(metrics, pd.DataFrame)
    assert "MAE" in metrics.columns
    assert "MSE" in metrics.columns
    assert "weight" in metrics.columns
    assert "cutoff_date" in metrics.columns
    assert len(metrics) == len(grouped_values)
    assert all(0 < metric < 1 for metric in metrics["MAE"])
    assert all(0 < metric < 1 for metric in metrics["MSE"])
    assert all(metric == 1 for metric in metrics["weight"])


def test_evaluate_metrics_on_forecasts_with_missing_values_grouped_by(forecasts_with_missing_values):
    metrics = evaluate_metrics(
        forecasts_with_missing_values,
        {"MAE": mean_absolute_error, "MSE": mean_squared_error},
        groupby="cutoff_date",
    )

    grouped_values = forecasts_with_missing_values["cutoff_date"].unique()

    assert isinstance(metrics, pd.DataFrame)
    assert "MAE" in metrics.columns
    assert "MSE" in metrics.columns
    assert "weight" in metrics.columns
    assert "cutoff_date" in metrics.columns
    assert len(metrics) == len(grouped_values)
    assert all(0 < metric < 1 for metric in metrics["MAE"])
    assert all(0 < metric < 1 for metric in metrics["MSE"])
    assert all(0 <= metric <= 1 for metric in metrics["weight"])
